unit UTree;

interface

uses
  UStack, SysUtils, ComCtrls;

type
  PNode = ^TNode;

  TTree = class
  private
    Root: PNode;
  protected
    procedure DestroyNode; // if we need to destroy only one node 
    constructor Create(ch: char); overload; // create tree as leaf; 
    constructor Create(ch: char; Aleft, Aright: TTree); overload;
    // create tree as branch 
    function IsLeaf: boolean; // checks if node is leaf 
    function GetValue: char; // getts data in root 
  public
    constructor Create; overload; // create empty tree; 
    destructor Destroy; override; // if we need to destroy all tree 
    procedure MakeFromStr(var s: string); // make tree from str; 
    function MakeToStr: string; // make tree to string 
    procedure MakeForView(Items: TTreeNodes); // Writes the tree to treeView 
    procedure ProcessTree; // simplifies the tree 
  end;

  TNode = record
    data: char;
    left, right: TTree;
  end;

implementation

const
  variables: set of char = ['a' .. 'z', 'A' .. 'Z'];
  signs: set of char = ['+', '-', '*', '/'];
  digits: set of char = ['0' .. '9'];

  { TTree }

constructor TTree.Create;    //create NULL tree
begin
  Root := nil;
end;

constructor TTree.Create(ch: char);   //create Tree as Leaf
begin
  new(Root);
  Root.data := ch;
  Root.left := nil;
  Root.right := nil;
end;

constructor TTree.Create(ch: char; Aleft, Aright: TTree);  //create Full tree
begin
  new(Root);
  Root.data := ch;
  Root.left := Aleft;
  Root.right := Aright;
end;

destructor TTree.Destroy; // if we need to destroy all tree 
begin
  if Root <> nil then
  begin
    if Root.left <> nil then
      Root.left.Destroy;
    if Root.right <> nil then
      Root.right.Destroy;
    dispose(Root);
  end;
  inherited Destroy;
end;

procedure TTree.DestroyNode; // if we need to destroy only one node 
begin
  if Root <> nil then
  begin
    Root.left := nil;
    Root.right := nil;
  end;
  Destroy;
end;

function TTree.GetValue: char; // getts data in root 
begin
  if Root.data in signs then
    result := #0
  else
    result := Root.data;
end;

function TTree.IsLeaf: boolean; // checks if node is leaf 
begin
  result := false;
  if Root.left = nil then
    if Root.right = nil then
      result := true;

end;

procedure TTree.MakeForView(Items: TTreeNodes);
// Writes the tree to treeView 

  procedure WriteNodes(var Root: PNode; Node: TTreeNode);
  var                              //recurrent procedure which write nodes
    Parent: TTreeNode;
  begin
    if Root.data in signs then
    begin
      Parent := Items.AddChild(Node, Root.data);
      WriteNodes(Root.left.Root, Parent);
      WriteNodes(Root.right.Root, Parent);
    end
    else
    begin
      Items.AddChild(Node, Root.data);
    end;
  end;

begin
  Items.Clear;          //don't forget to clear Items
  WriteNodes(Root, nil);
end;

procedure TTree.MakeFromStr(var s: string); // make tree from string 
var
  i, len: integer;
  St: TStack;
  left, right, Node: TTree;

  procedure SkipSpaces(var s: string; var i: integer); // skip all spaces 
  var
    len: integer;
  begin
    len := length(s);
    while (i <= len) and (s[i] = ' ') do
      inc(i);
  end;

begin
  s := Trim(s);
  len := length(s);
  St := TStack.Create;
  i := 1;
  while i <= len do
  begin
    SkipSpaces(s, i);          //skip spaces
    if s[i] in signs then      //and see what char appeared
    begin
      right := St.Pop;         //if it's ariphmetical sign
      left := St.Pop;          //then make full tree
      Node := TTree.Create(s[i], left, right);
      St.Push(Node);
    end
    else if (s[i] in (digits + variables)) then
    begin
      Node := TTree.Create(s[i]);  //otherwise we make tree as leaf
      St.Push(Node);
    end;
    inc(i);
  end;
  Node := St.Pop;      //at the end we take the last tree
  Root := Node.Root;     //and that will be final result
  Node.Root := nil;
  Node.DestroyNode; // and destroy pointer to result 
end;

function TTree.MakeToStr: string; // make tree to string 
var
  rez: string;           //it's recurrent function too
  ch: char;
begin
  rez := '';                      //it's everything easy here ;-)
  if Root.left <> nil then
    rez := '(' + Root.left.MakeToStr + ')';
  ch := Root.data;
  rez := rez + ch;
  if Root.right <> nil then
    rez := rez + '(' + Root.right.MakeToStr + ')';
  result := rez;
end;

procedure TTree.ProcessTree; // simplifies the tree 
var
  ch1, ch2: char;

  procedure GetRightTree;   //if we need to take right subtree
  var
    buf: TTree;
  begin
    buf := Root.right;
    Root.left.Destroy;
    dispose(Root);
    Root := buf.Root;
    buf.Root := nil;
    buf.DestroyNode;
  end;

  procedure GetLeftTree;    //if we need to take left subTree
  var
    buf: TTree;
  begin
    buf := Root.left;
    Root.right.Destroy;
    dispose(Root);
    Root := buf.Root;
    buf.Root := nil;
    buf.DestroyNode;
  end;

  procedure MakeZeroTree;   //if we need to make zero leaf
  var
    buf: TTree;
  begin
    Root.left.Destroy;
    Root.left := nil;
    Root.right.Destroy;
    Root.right := nil;
    Root.data := '0';
  end;

begin                             //as you noticed, this procedure is too recurrent :-)
  if Root.left <> nil then
    Root.left.ProcessTree; // at the very beginning 
  if Root.right <> nil then
    Root.right.ProcessTree; // see subtrees

  if Root.data = '+' then             //check all situations
  begin
    if Root.left.GetValue = '0' then
      GetRightTree else
    if Root.right.GetValue = '0' then
      GetLeftTree;
  end else

  if Root.data = '-' then
  begin
    if Root.right.GetValue = '0' then
      GetLeftTree 
    else
    begin

      ch1 := Root.left.GetValue;
      ch2 := Root.right.GetValue;
      if (ch1 = ch2) and (ch1 <> #0) then
        MakeZeroTree;
    end;
  end else

  if Root.data = '/' then
  begin
    if Root.left.GetValue = '0' then
    begin
      MakeZeroTree;
    end;
  end else

  if Root.data = '*' then
  begin
    ch1 := Root.left.GetValue;
    ch2 := Root.right.GetValue;
    if (ch1 = '0') or (ch2 = '0') then
      MakeZeroTree
    else if ch1 = '1' then
      GetRightTree
    else if ch2 = '1' then
      GetLeftTree;

  end;

end;

end.
